@@ -3,7 +3,7 @@ package VCS::CMSynergy;
 # Copyright (c) 2001-2013 argumentum GmbH
 # See COPYRIGHT section in VCS/CMSynergy.pod for usage and distribution rights.
 
-our $VERSION = '1.40';
+our $VERSION = '1.41';
 
 use 5.006_000;				# i.e. v5.6.0
 use strict;
@@ -478,8 +478,8 @@ sub _query
 	next unless length($_);			# skip empty leading record
 
 	my @cols = split(/\Q$FS\E/, $_, -1);	# don't strip empty trailing fields
-	my %finduse;
 
+	my %finduse;
 	if ($want_finduse)
 	{
 	    # finduse information is the last "column" 
@@ -1182,30 +1182,7 @@ sub relations_arrayref
     my ($self, %args) = @_;
     validate([\%args], Dict[ _FROM_TO_NAME ]);
 
-    my ($rc, $out, $err) = $self->_ccm(qw/ relate -show -nf -l /,
-	    map { defined $args{$_} ? ( "-$_" => $args{$_}) : () } qw/from to name/);
-
-    # NOTE: If there are no hits, `ccm relate' exits 
-    # with status 1 (classic mode) or 6 (web mode),
-    # but produces no output on either stdout and stderr.
-    return [ ] if $rc != 0 and $out eq "" and $err eq "";
-    return $self->set_error($err || $out) unless $rc == 0;
-
-    my (@result, $from, $to);
-    foreach (split(/\n/, $out))	        	# split into lines 
-    {
-        # NOTE: In web mode, the "default long format" (option "-l") prints
-        #   (from-info) relation-name (to-info) timestamp
-        # where info is
-        #    objectname status owner
-        # which can be reasonably parsed assuming that status and owner
-        # don't contain blanks.
-        my ($from, $name, $to, $timestamp) = /^\((.*?)\) (\S+) \((.*?)\) (.*)$/;
-        ($from, $to) = map { s/ \S+ \S+$//; $self->object($_) } $from, $to;
-        push @result, [$from, $name, $to, $timestamp];
-    }
-
-    return \@result;
+    return _flatten_rows($self->relations_hashref(%args), [qw/ from name to create_time /]);
 }
 
 
@@ -1217,13 +1194,14 @@ sub relations_hashref
                              to_attributes   => Optional[ArrayRef[Str]] ]);
 
     my %defaulted;
-    foreach (qw/from_attributes to_attributes/)
+    foreach my $arg (qw/from to/)
     {
-	next if $args{$_};
+        my $attrs = "${arg}_attributes";
+        next if $args{$attrs};
 
         # default to keyword "objectname"
-        $args{$_} = [ qw/objectname/ ];
-        $defaulted{$_}++;
+        $args{$attrs} = [ qw/objectname/ ];
+        $defaulted{$arg}++;
     }
 
     my $result = $self->_relations(\%args, ROW_HASH);
@@ -1233,7 +1211,7 @@ sub relations_hashref
     # hash containing the sole key "objectname" with its value
     foreach my $arg (qw/from to/)
     {
-	if ($defaulted{"${arg}_attributes"})
+	if ($defaulted{$arg})
 	{
 	    $_->{$arg} = $_->{$arg}->{objectname} foreach @$result;
 	}
@@ -1597,6 +1575,50 @@ sub list_attributes
 }
 
 
+sub properties_hashref
+{
+    my $self = shift;
+    my ($file_specs, $keywords) = validate(\@_, ArrayRef, _KEYWORDS);
+    # FIXME ArrayRef[_FILE_SPEC]  ???
+
+    return $self->_properties($file_specs, $keywords, ROW_HASH);
+}
+
+sub properties_object
+{
+    my $self = shift;
+    my ($file_specs, $keywords) = validate(\@_, ArrayRef, _KEYWORDS);
+    # FIXME ArrayRef[_FILE_SPEC]  ???
+
+    return $self->_properties($file_specs, $keywords, ROW_OBJECT);
+}
+
+sub _properties
+{
+    my ($self, $file_specs, $keywords, $row_type) = @_;
+    # NOTE: only ROW_HASH and ROW_OBJECT are allowed for $row_type
+
+    return [ ] if @$file_specs == 0;            # silly, but consistent
+
+    my $want = _want($row_type, $keywords);
+    my $format = $RS . join($FS, values %$want) . $FS;
+
+    my ($rc, $out, $err) = 
+	$self->_ccm(qw/properties -nf -format/, $format, @$file_specs);
+    return $self->set_error($err || $out) unless $rc == 0;
+
+    my @result;
+    foreach (split(/\Q$RS\E/, $out))		# split into records 
+    {
+	next unless length($_);			# skip empty leading record
+
+	my @cols = split(/\Q$FS\E/, $_, -1);	# don't strip empty trailing fields
+	push @result, $self->_query_result($want, \@cols, $row_type);
+    }
+    return \@result;
+}
+
+
 sub property
 {
     my $self = shift;
